import tensorflow as tf
from tensorflow import keras as ks
import matplotlib.pyplot as plt
import numpy as np


class Model:

    """
    A class used to contain a Keras Model

    ...

    Attributes
    ----------

    Methods
    -------
    predict(tensor)
        Predict the matrix of inputs given by the tensor
    get_model_structure()
        Returns an instance of ModelStructure with the data of the model structure.
    get_name_layer(k:int)
        Returns the name of the layer k.
    evaluate(x,y)
        Returns a list with the loss and the accuracy for the tuple (x,y)
    """

    def __init__(self, model_keras: tf.keras.Model, model_path: str):
        """
        Parameters
        ----------
        model_keras : tf.keras.Model
            a keras model
        model_path : str
            a path to load a keras model
        """

        if(model_keras != None):
            self.model_keras = model_keras
        else:
            self.model_keras = tf.keras.models.load_model(model_path)

        self.layers_outpus = [
            layer.output for layer in self.model_keras.layers]
        self.layers_names = [layer.name for layer in self.layers_outpus]
        self.layers_names = [
            name[:name.index('/')]for name in self.layers_names]

        self.model_layers = ks.models.Model(
            inputs=self.model_keras.input, outputs=self.layers_outpus)

        self.model_structure = ModelStructure(self.model_keras)

    def predict(self, tensor: tf.Tensor):
        """Predict a matrix of inputs given by the tensor

        Parameters
        ----------
        tensor:tf.Tensor

        Returns
        -------
        x:X
            Returns an object that indexes a three dimensional list
            through a tuple (i,j,l) where i represents the row and j
            represents the column corresponding to the input matrix
            and l represents the l-layer of a model. The values in the
            list are the activations values in the layer l for the
            input located in the row i and the column j.
        """

        aux = []
        aux.append(tensor.shape[0]*tensor.shape[1])
        for i in range(len(tensor.shape)-2):
            aux.append(tensor.shape[i+2])

        x = tf.reshape(tensor, aux)

        x = self.model_layers.predict(x)

        return Block(x, tensor.shape[0], tensor.shape[1], self.model_structure)

    def get_model_structure(self):
        """

        Parameters
        ----------

        Returns
        -------
        x:ModelStructure

        """
        return self.model_structure

    def get_name_layer(self, k):
        """

        Parameters
        ----------
        k:int
            the index of the layer
        Returns
        -------
        x:str
            Returns the name of the layer k.

        """
        return self.layers_names[k]

    def evaluate(self, x, y):
        """

        Parameters
        ----------
        x
            the input of the model
        y
            the desired output for x


        Returns
        -------
        result:list
            Returns a list with the loss and the accuracy for the tuple (x,y)

        """
        return self.model_keras.evaluate(x=x, y=y)


class ModelStructure:
    """
    A class used to represent the model structure
    ...

    Attributes
    ----------
    layers:list
        a list with the layer structure for each layer
    number_layers:int
        the number of layers in the model
    map_layers_features:list
        a list containing a map between the layers and the layers with features
    layers_features:list
        a list with the layer structure for each layer with features
    number_layers_features:int
        the number of layers with features
    Methods
    -------

    """

    def __init__(self, model: tf.keras.Model = None):
        """
        Parameters
        ----------
        model : tf.keras.Model
            a keras model
        """
        self.layers = []
        self.number_layers = 0
        self.map_layers_features = []
        self.layers_features = []
        self.number_layers_features = 0

        if model != None:

            self.layers_outpus = [layer.output for layer in model.layers]
            k = 0
            self.number_layers_features = 0
            for layer in self.layers_outpus:

                shape = layer.shape[1:]
                layer_structure = LayerStructure()
                layer_structure.shape = shape
                h = 1
                for i in shape:
                    h = h*i
                layer_structure.number_activations = h

                if len(shape) > 2:
                    layer_structure.number_features = shape[-1]
                    layer_structure.shape_feature = shape[0:-1]
                    self.layers_features.append(layer_structure)
                    self.map_layers_features.append(k)
                    self.number_layers_features += 1
                else:
                    layer_structure.number_features = 0
                    pass

                self.layers.append(layer_structure)
                k = k+1

            self.number_layers = k


class LayerStructure:
    """
    A class used to represent the structure of a layer in a keras model 
    ...

    Attributes
    ----------
    number_features:int
        the number of features in the layer
    shape_feature:tuple
        the shape for the features in the layer
    number_activations
        the number of activations in the layer
    shape:tuple
        the shape of the layer

    Methods
    -------

    """

    def __init__(self):
        self.number_features = None
        self.shape_feature = None
        self.number_activations = None
        self.shape = None


class Block:
    """
    A class used to represent the output when the model predicts a block of inputs. The block can be indexed 
    by three or four indexes to get a particular output. Besides, the block has some information of the 
    dataset from which the inputs come. 
    ...

    Attributes
    ----------
    height:int
        the height of the block
    width:int
        the width of the block
    model_structure:ModelStructure
        the model structure

    Methods
    -------
    set_rows_dataset(rows:list)

    set_columns_dataset(columns:list)

    set_state_transpose_dataset(state:bool)

    get_rows_dataset()

    get_columns_dataset()

    is_dataset_transpose()

    """

    def __init__(self, x, height: int, width: int, model_structure: ModelStructure):
        self.x = x
        self.height = height
        self.width = width
        self.model_structure = model_structure
        self.rows_dataset = None
        self.columns_dataset = None
        self.state_transpose_dataset = None

    def __getitem__(self, tuple):
        # tuple[0]=i,tuple[1]=j,tuple[2]=l,tuple[3]=f

        if len(tuple) == 3:
            layer = self.x[tuple[2]][tuple[0]*self.width+tuple[1]]
            return layer

        if len(tuple) == 4:
            k = self.model_structure.map_layers_features[tuple[2]]
            layer = self.x[k][tuple[0]*self.width+tuple[1]]
            feature = tf.unstack(layer, axis=tf.rank(layer)-1)
            feature = feature[tuple[3]]
            return feature

        return None

    def __setitem__(self, tuple, value):

        if len(tuple) == 3:
            self.x[tuple[2]][tuple[0]*self.width+tuple[1]] = value

        if len(tuple) == 4:
            layer = self.x[tuple[2]][tuple[0]*self.width+tuple[1]]
            axis = tf.rank(layer)-1
            layer_aux = tf.unstack(layer, axis=axis)
            layer_aux[tuple[3]] = value
            self.x[tuple[2]][tuple[0]*self.width+tuple[1]
                             ] = tf.stack(layer_aux, axis=axis)

    def set_rows_dataset(self, rows: list):
        self.rows_dataset = rows

    def set_columns_dataset(self, columns: list):
        self.columns_dataset = columns

    def set_state_transpose_dataset(self, state: bool):
        self.state_transpose_dataset = state

    def get_rows_dataset(self):
        return self.rows_dataset

    def get_columns_dataset(self):
        return self.columns_dataset

    def is_dataset_transpose(self):
        return self.state_transpose_dataset



import tensorflow as tf
from tensorflow import keras as ks
import numpy as np
from model import Model
from dataset import DataSet
from tqdm import tqdm


class Iterator:

    """
    A class used to iterate over a Model and a DataSet 2D

    ...

    Attributes
    ----------

    Methods
    -------
    get_model()
        Returns the model.
    get_block(width:int,height:int)
        Returns the block iterator.


    """

    def __init__(self, model: Model, dataset: DataSet):
        """
        Parameters
        ----------
        model: Model
        dataset : DataSet
        """

        self.dataset = dataset
        self.model = model
        self.n = self.dataset.get_height()
        self.m = self.dataset.get_width()

    def get_model(self):
        """
        Returns
        -------
        model:Model
            Returns the model

        """
        return self.model

    def get_block(self, height: int, width: int):
        """

        Parameters
        ----------
        height:int
        widht:int

        Returns
        -------
        x:Iterator
            Returns a "vertical" iterator where each item in this
            iterator is a tuple (height,blocks_iterator) where 
            blocks_iterator is an "horizontal" iterator in which 
            each item is a tuple (width,block_activations) where 
            block_activations is an object that indexes a three 
            dimensional list through a tuple (i,j,l) where i 
            represents the row and j represents the column 
            corresponding to the block and l represents the l-layer
            of a model. The values in the list are the activations 
            values in the layer l for the input located in the row i 
            and the column j, and width is the width of the block and 
            the height of the block is the height returned previously.
        """
        batchs = self.generate_batchs(height, self.n)
        for batch in tqdm(batchs):
            yield len(batch), self.get_block_activations(batch, width)

    def get_block_activations(self, batch_vertical, width):
        batchs = self.generate_batchs(width, self.m)
        for batch in batchs:
            x = self.dataset.get_matrix(batch_vertical, batch)
            bloque = self.model.predict(x)
            bloque.set_rows_dataset(batch_vertical)
            bloque.set_columns_dataset(batch)
            bloque.set_state_transpose_dataset(self.dataset.is_transpose())
            yield len(batch), bloque

    def generate_batchs(self, size_batch, n):
        aux = range(n)
        h = int(n/size_batch)
        batchs = [aux[j*size_batch:(j+1)*size_batch] for j in range(h)]
        if(n % size_batch > 0):
            batchs.append(aux[h*size_batch:n])
            h = h+1
        return batchs



import tensorflow as tf
from iterator import Iterator
from time import time
from variance import Variance
from metric import Metric
from transformationsdataset import TransformationsDataSet
from variance import VarianceCalculations


class SameEquivariance:

    def __init__(self, iterator: Iterator = None, dataset: TransformationsDataSet = None):
        if iterator != None:
            self.set_iterator(iterator)
        if dataset != None:
            self.set_dataset(dataset)

        self.result = None

    # overriding abstract method
    def set_iterator(self, iterator: Iterator):
        self.iterator = iterator
        self.model_structure = self.iterator.get_model().model_structure

    def set_dataset(self, dataset: TransformationsDataSet):
        self.dataset = dataset

    # overriding abstract method
    def get_result(self):
        return self.result

    # overriding abstract method
    def compute(self, height, width):

        initial_time = time()
        calculations = VarianceCalculations(self.model_structure)

        for memory, blocks in self.iterator.get_block(height, width):
            calculations.renew(memory)
            for length_mov, block in blocks:

                if block.is_dataset_transpose() == False:

                    for k in self.model_structure.map_layers_features:
                        j = 0
                        for t in block.get_columns_dataset():
                            for i in range(block.height):
                                for f in range(self.model_structure.layers[k].number_features):
                                    aux = block[(i, j, k, f)]
                                    block[(i, j, k, f)] = self.dataset.apply_inverse_transformation(
                                        block[(i, j, k, f)], t)
                            j = j+1
                else:
                    for k in self.model_structure.map_layers_features:
                        j = 0
                        for t in block.get_rows_dataset():
                            for i in range(block.width):
                                for f in range(self.model_structure.layers[k].number_features):
                                    aux = block[(j, i, k, f)]
                                    block[(j, i, k, f)] = self.dataset.apply_inverse_transformation(
                                        block[(j, i, k, f)], t)
                            j = j+1

                calculations.add(block, length_mov)
            calculations.update()

        calculations.finish()

        self.variance_layers = calculations.variance_layers
        self.variance_layers_activations = calculations.variance_layers_activations

        self.variance_layers_features = []
        aux = []
        for k in self.model_structure.map_layers_features:

            layer = self.variance_layers_activations[k]
            variance_features = []

            for f in range(self.model_structure.layers[k].number_features):

                feature = tf.unstack(layer, axis=tf.rank(layer)-1)[f]
                variance_features.append(tf.reduce_mean(feature))

            self.variance_layers_features.append(variance_features)
            aux.append(tf.reduce_mean(tf.stack(variance_features)))

        self.time = time() - initial_time

        self.total_variance = tf.reduce_mean(tf.stack(aux))
        self.result = (self.total_variance, self.variance_layers_features,
                       self.variance_layers_activations)
        return self.result



import tensorflow as tf
from iterator import Iterator
from time import time
from variance import Variance
from metric import Metric


class FeaturesVariance(Metric):

    def __init__(self, iterator: Iterator = None):
        if iterator != None:
            self.set_iterator(iterator)

        self.result = None

    # overriding abstract method
    def set_iterator(self, iterator: Iterator):
        self.iterator = iterator
        self.variance = Variance(iterator)
        self.model_structure = self.iterator.get_model().model_structure

    # overriding abstract method
    def get_result(self):
        return self.result

    # overriding abstract method
    def compute(self, height, width):
        self.variance.compute(height, width)
        self.variance_layers_activations = self.variance.variance_layers_activations

        self.variance_layers_features = []
        aux = []
        for k in self.model_structure.map_layers_features:

            layer = self.variance_layers_activations[k]
            variance_features = []

            for f in range(self.model_structure.layers[k].number_features):

                feature = tf.unstack(layer, axis=tf.rank(layer)-1)[f]
                variance_features.append(tf.reduce_mean(feature))

            aux.append(tf.reduce_mean(tf.stack(variance_features)))

            self.variance_layers_features.append(variance_features)

        self.total_variance = tf.reduce_mean(tf.stack(aux))
        self.result = (self.total_variance, self.variance_layers_features,
                       self.variance_layers_activations)
        return self.result

    def get_feature(self, layer, feature):
        layer = self.variance_layers_activations[layer]
        feature = tf.unstack(layer, axis=tf.rank(layer)-1)[feature]
        return feature



from abc import ABC, abstractmethod
from iterator import Iterator


class Metric(ABC):
    """
    An abstract class thats represents the main behavior of a Metric
    ...

    Attributes
    ----------

    Methods
    -------
    compute(height:int,width:int)
        computes the metric iterating over a block with such height and width
    set_iterator(iterator:Iterator)
        set the iterator in the metric
    get_result()
        returns the results computed by the metric

    """

    # abstract method
    def compute(self, height, width):
        pass

    # abstract method
    def set_iterator(self, iterator: Iterator):
        pass

    # abstract method
    def get_result(self):
        pass



import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np


class Graphs:
    """
    A class used to graph heatmaps
    ...

    Attributes
    ----------

    Methods
    -------
    hetmap_list2D(list2D:list,columns_name:list,title:str)

    heatmap_matrix(matrix,columns_name:list,title:str)

    """

    def __init__(self):
        pass

    def plot(self, x, y, title=None):
        plt.figure()
        plt.title(title)
        plt.plot(x, y)

    def heatmap_list2D(self, list2D: list, columns_names: list = None, title: str = None):
        """It Graphs a heatmap  

        Parameters
        ----------
        list2D:list
            a 2D list with the heatmap values
        columns_name:list
            it contains the names for each column in the heat map
        title:str
            the heatmap title

        """

        m = len(list2D)
        y = [tf.reduce_max(list2D[i]) for i in range(m)]
        vmax = tf.reduce_max(tf.stack(y))
        y = [tf.reduce_min(list2D[i]) for i in range(m)]
        vmin = tf.reduce_min(tf.stack(y))
        if(vmin < 0):
            vmin = 0

        f, axes = plt.subplots(1, m)
        for i in range(m):
            ax = axes[i]
            ax.axis("off")
            column = np.array(list2D[:][i])
            column = column[:, np.newaxis]

            mappable = ax.imshow(column, vmax=vmax, vmin=vmin,
                                 cmap='inferno', aspect="auto")

            if columns_names != None:
                name = columns_names[i]
            else:
                name = ""

            if len(name) > 7:
                name = name[:10]+"."

            ax.set_title(name, fontsize=8, rotation=45)

        f.subplots_adjust(right=0.8)
        cbar_ax = f.add_axes([0.85, 0.15, 0.05, 0.7])
        cbar = f.colorbar(mappable, cax=cbar_ax, extend='max')
        cbar.cmap.set_over('green')
        cbar.cmap.set_bad(color='blue')
        if title != None:
            f.suptitle(title)

    def heatmap_matrix(self, matrix, columns_names: list = None, title: str = None):
        """It Graphs a heatmap  
        Parameters
        ----------
        matrix
            a 2D array representing a matrix with the heatmap values
        columns_name:list
            it contains the names for each column in the heat map
        title:str
            the heatmap title
            """

        matrix = tf.convert_to_tensor(matrix)
        shape = matrix.shape
        n = shape[0]
        m = shape[1]

        vmax = tf.reduce_max(matrix)
        vmin = tf.reduce_min(matrix)
        if(vmin < 0):
            vmin = 0

        f, axes = plt.subplots(1, m)
        for i in range(m):
            ax = axes[i]
            ax.axis("off")
            column = np.array(matrix[:, i])
            column = column[:, np.newaxis]

            mappable = ax.imshow(column, vmax=vmax, vmin=vmin,
                                 cmap='inferno', aspect="auto")

            if columns_names != None:
                name = columns_names[i]
            else:
                name = ""

            if len(name) > 7:
                name = name[:10]+"."

            ax.set_title(name, fontsize=8, rotation=45)

        f.subplots_adjust(right=0.8)
        cbar_ax = f.add_axes([0.85, 0.15, 0.05, 0.7])
        cbar = f.colorbar(mappable, cax=cbar_ax, extend='max')
        cbar.cmap.set_over('green')
        cbar.cmap.set_bad(color='blue')

        if title != None:
            f.suptitle(title)



import tensorflow as tf
from iterator import Iterator
from time import time
from metric import Metric


class Variance(Metric):

    def __init__(self, iterator: Iterator = None):
        self.iterator = iterator
        self.result = None

    # overriding abstract method
    def set_iterator(self, iterator: Iterator):
        self.iterator = iterator

    # overriding abstract method
    def get_result(self):
        return self.result

    # overriding abstract method
    def compute(self, height, width):

        initial_time = time()
        calculations = VarianceCalculations(
            self.iterator.get_model().model_structure)

        for memory, blocks in self.iterator.get_block(height, width):
            calculations.renew(memory)
            for length_mov, block in blocks:
                calculations.add(block, length_mov)
            calculations.update()

        calculations.finish()

        self.variance_layers = calculations.variance_layers
        self.variance_layers_activations = calculations.variance_layers_activations
        self.total_variance = tf.reduce_mean(tf.stack(self.variance_layers))

        self.result = (self.total_variance, self.variance_layers,
                       self.variance_layers_activations)

        self.time = time() - initial_time

        return self.result


class VarianceCalculations:

    def __init__(self, model_structure):
        self.model_structure = model_structure
        self.divisor_number = 0
        self.variance_layers_activations = [tf.zeros(
            shape=self.model_structure.layers[k].shape, dtype=tf.dtypes.float32) for k in range(self.model_structure.number_layers)]

    def renew(self, memory_length):
        self.count = 0
        self.memory_length = memory_length
        self.mean = [[tf.zeros(shape=self.model_structure.layers[k].shape, dtype=tf.dtypes.float32)
                      for k in range(self.model_structure.number_layers)] for q in range(memory_length)]
        self.moment = [[tf.zeros(shape=self.model_structure.layers[k].shape, dtype=tf.dtypes.float32)
                        for k in range(self.model_structure.number_layers)] for q in range(memory_length)]

        self.divisor_number += memory_length

    def add(self, block, mov_length):
        self.mov_length = mov_length
        for q in range(self.mov_length):
            self.count += 1
            for r in range(self.memory_length):

                for k in range(self.model_structure.number_layers):

                    layer = block[(r, q, k)]
                    last_mean = self.mean[r][k]
                    self.mean[r][k] = tf.add(
                        last_mean, tf.subtract(layer, last_mean)/self.count)
                    self.moment[r][k] = tf.add(self.moment[r][k], tf.multiply(
                        tf.subtract(layer, last_mean), tf.subtract(layer, self.mean[r][k])))

    def update(self):
        for r in range(self.memory_length):
            self.variance_layers_activations = [tf.add(self.variance_layers_activations[k], (
                self.moment[r][k]/(self.count-1))) for k in range(self.model_structure.number_layers)]

    def finish(self):
        self.variance_layers_activations = [self.variance_layers_activations[k] /
                                            self.divisor_number for k in range(self.model_structure.number_layers)]
        self.variance_layers = [tf.reduce_mean(
            self.variance_layers_activations[k]) for k in range(self.model_structure.number_layers)]



from model import Model
from transformationsdataset import TransformationsDataSet
from metric import Metric
from iterator import Iterator
import matplotlib.pyplot as plt


class Analyzer:

    """
    A class that analyzes the metric results vs the accuracy

    ...

    Attributes
    ----------

    Methods
    -------
    run(n:int)



    """

    def __init__(self, model: Model, dataset: TransformationsDataSet, metric: Metric):
        """
        Parameters
        ----------
        model: Model
        dataset : DataSet
        metric: Metric
        """

        self.model = model
        self.dataset = dataset
        self.metric = metric

    def run(self, n):
        """
        Parameters
        ----------
        n: int
            the amount of transformations in each group
        """

        self.x = []
        self.y = []

        for i in range(self.dataset.get_size_transformations_groups()):
            self.dataset.set_transformation_group(i, n)
            iterator = Iterator(self.model, self.dataset)
            self.metric.set_iterator(iterator)
            var = self.metric.compute(10, 5)[0]
            self.x.append(var)
            w = self.dataset.get_testing_data()
            x1 = w[0]
            y1 = w[1]

            acc = self.model.evaluate(x1, y1)[1]
            self.y.append(acc)

        plt.figure()
        plt.plot(self.x, self.y, 'ro')



import tensorflow as tf
from tensorflow import keras as ks
import matplotlib.pyplot as plt
import numpy as np
from model import Model
from mnistdataset import MnistDataSet
from iterator import Iterator
from variance import Variance

path = ""


model = Model(path+"/modelosimple.h5")
dataset = MnistDataSet()
iterator = Iterator(model, dataset)

variance = Variance(iterator)

variance.compute(10, 2)

plt.figure()
plt.title("Transformational Variance")
plt.plot(model.layers_names, variance.variance_layers)
print(variance.variance_layers)

model = Model(path+"/modelosimple.h5")
dataset = DataSet()
dataset.transpose()
iterator = Iterator(model, dataset)

variance = Variance(iterator)

variance.compute(10, 2)

plt.figure()
plt.title("Muestral Variance")
plt.plot(model.layers_names, variance.variance_layers)
print(variance.variance_layers)



import random
import tensorflow as tf
from tensorflow import keras as ks
import matplotlib.pyplot as plt
import numpy as np
from abc import ABC, abstractmethod
from transformationsdataset import TransformationsDataSet


class Transformation(ABC):
    """
    A class used to contain a Keras Model

    ...

    Attributes
    ----------

    Methods
    -------
    predict(tensor)
        Predict the matrix of inputs given by the tensor

    """

    # abstract method
    def apply(self, tensor):
        pass

    # abstract method
    def apply_inverse(self, tensor):
        pass


class AffineTransformation(Transformation):

    def __init__(self, theta=0, tx=0, ty=0, shear=0, zx=1, zy=1):

        self.data = ks.datasets.mnist.load_data()
        self.theta = theta
        self.tx = tx
        self.ty = ty
        self.shear = shear
        self.zx = zx
        self.zy = zy

    # overriding abstract method
    def apply(self, tensor):
        return tf.keras.preprocessing.image.apply_affine_transform(tensor, theta=self.theta, tx=self.tx, ty=self.ty, shear=self.shear, zx=self.zx, zy=self.zy)

    # overriding abstract method
    def apply_inverse(self, tensor):
        return tf.keras.preprocessing.image.apply_affine_transform(tensor, theta=-self.theta, tx=-self.tx, ty=-self.ty, shear=-self.shear, zx=-self.zx, zy=-self.zy)


class MnistTransformationsDataSet(TransformationsDataSet):

    def __init__(self):

        self.data = ks.datasets.mnist.load_data()
        self.results = self.data[1][1]
        self.samples = self.data[1][0]
        self.samples = self.samples.reshape(-1, 28, 28, 1)
        self.height = self.samples.shape[0]
        self.mean = np.mean(self.data[0][0])
        self.std = np.std(self.data[0][0])
        # Samples Normalization
        for i in range(self.height):
            self.samples[i, :, :] = (self.samples[i, :, :]-self.mean)/self.std

        self.original_samples = self.samples
        self.original_results = self.results

        self.shape = self.samples.shape

        self.transformations = [AffineTransformation()]
        self.width = len(self.transformations)
        self.matrix_transpose = False
        self.size_transformations_groups = 7

    # overriding abstract method
    def get_data_shape(self):
        return self.shape

    # overriding abstract method
    def get_width(self):
        return self.width

    # overriding abstract method
    def get_height(self):
        return self.height

    # overriding abstract method
    def get_matrix(self, rows, columns):
        m = []

        if(self.matrix_transpose == False):
            for i in rows:
                x = [self.transformations[j].apply(tf.make_ndarray(
                    tf.make_tensor_proto(self.samples[i]))) for j in columns]
                m.append(x)
        else:
            for i in rows:
                x = [self.transformations[i].apply(tf.make_ndarray(
                    tf.make_tensor_proto(self.samples[j]))) for j in columns]
                m.append(x)

        return tf.convert_to_tensor(m)

    # overriding abstract method
    def transpose(self):
        if(self.matrix_transpose == True):
            self.matrix_transpose = False
        else:
            self.matrix_transpose = True
        aux = self.height
        self.height = self.width
        self.width = aux

    # overriding abstract method
    def apply_inverse_transformation(self, tensor: tf.Tensor, index_transformation):
        if len(tensor.shape) == 2:
            shape_aux = tensor.shape
            shape = (tensor.shape[0], tensor.shape[1], 1)
            tensor = tf.reshape(tensor, shape)
            array = tf.make_ndarray(tf.make_tensor_proto(tensor))
            img = self.transformations[index_transformation].apply_inverse(
                array)
            img = tf.reshape(img, shape_aux)
            return img
        else:
            return tensor

    # overriding abstract method
    def get_testing_data(self):

        x = []
        y = []

        for i in range(self.samples.shape[0]):
            for t in self.transformations:
                x.append(t.apply(tf.make_ndarray(
                    tf.make_tensor_proto(self.samples[i]))))
                y.append(self.results[i])

        x = tf.stack(x)
        y = tf.stack(y)
        return (x, y)

    def select_random_samples(self, n: int):
        indexes = random.sample(range(self.original_samples.shape[0]), n)
        self.select_samples(indexes)

    def select_samples(self, indexes):
        aux_list = []
        self.results = []
        n = 0
        for i in indexes:
            aux_list.append(self.original_samples[i, :, :, :])
            self.results.append(self.original_results[i])
            n = n+1
        aux_list = tf.stack(aux_list, axis=0)
        self.samples = aux_list

        if self.matrix_transpose == False:
            self.height = n
        else:
            self.width = n

    # overriding abstract method
    def is_transpose(self):
        return self.matrix_transpose

    # overriding abstract method
    def get_size_transformations_groups(self):
        return self.size_transformations_groups

    # overriding abstract method
    def set_transformation_group(self, group, n):

        self.transformations = []
        if group == 0:
            for i in range(n):
                self.transformations.append(AffineTransformation())
        if group == 1:
            h = int(360/n)
            for i in range(n):
                self.transformations.append(AffineTransformation(theta=i*h))
        elif group == 2:
            width = self.original_samples[0].shape[1]
            if n <= width:
                h = int(width/n)
                for i in range(n):
                    self.transformations.append(AffineTransformation(tx=i*h))
            else:
                for i in range(n):
                    self.transformations.append(AffineTransformation(tx=i))
        elif group == 3:
            height = self.original_samples[0].shape[0]
            if n <= height:
                h = int(height/n)
                for i in range(n):
                    self.transformations.append(AffineTransformation(ty=i*h))
            else:
                for i in range(n):
                    self.transformations.append(AffineTransformation(ty=i))
        elif group == 4:
            n1 = int(n/2)
            n2 = n1
            if n != (n1+n2):
                n1 = n1+1

            height = self.original_samples[0].shape[0]
            if n1 <= height:
                h = int(height/n1)
                for i in range(n1):
                    self.transformations.append(AffineTransformation(ty=i*h))
            else:
                for i in range(n1):
                    self.transformations.append(AffineTransformation(ty=i))

            width = self.original_samples[0].shape[1]
            if n2 <= width:
                h = int(width/n2)
                for i in range(n2):
                    self.transformations.append(AffineTransformation(tx=i*h))
            else:
                for i in range(n2):
                    self.transformations.append(AffineTransformation(tx=i))

        elif group == 5:
            n1 = int(n/2)
            n2 = n1
            if n != (n1+n2):
                n1 = n1+1

            h = int(360/n1)
            for i in range(n1):
                self.transformations.append(AffineTransformation(theta=i*h))

            height = self.original_samples[0].shape[0]
            if n2 <= height:
                h = int(height/n2)
                for i in range(n2):
                    self.transformations.append(AffineTransformation(ty=i*h))
            else:
                for i in range(n2):
                    self.transformations.append(AffineTransformation(ty=i))

        elif group == 6:
            n1 = int(n/2)
            n2 = n1
            if n != (n1+n2):
                n1 = n1+1

            h = int(360/n1)
            for i in range(n1):
                self.transformations.append(AffineTransformation(theta=i*h))

            width = self.original_samples[0].shape[1]
            if n2 <= width:
                h = int(width/n2)
                for i in range(n2):
                    self.transformations.append(AffineTransformation(tx=i*h))
            else:
                for i in range(n2):
                    self.transformations.append(AffineTransformation(tx=i))

        else:
            pass

        self.transformation_group = group

        if self.matrix_transpose == False:
            self.width = n
        else:
            self.height = n



from setuptools import find_packages, setup, Command
from shutil import rmtree
import sys
import os
import io
from pathlib import Path

from os import path
this_directory = os.path.abspath(os.path.dirname(__file__))

with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


url="https://github.com/midusi/transformational_measures_tf/"
VERSION="0.4alpha"

class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        dist_path=Path(this_directory)/'dist'
        if dist_path.exists():
            self.status('Removing previous builds…')
            rmtree(dist_path)

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(
            '{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')

        sys.exit()


setup(
    name="transformational_measures_tf",
    version=VERSION,
    python_requires='>=3.6',
    packages=find_packages(),
    scripts=[],

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires=[
        "numpy",
        "scipy",
        "keras",
        "matplotlib",
        "tqdm",
    ],

    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.txt", "*.rst"],
        # And include any *.msg files found in the "hello" package, too:
        "hello": ["*.msg"],
    },
    zip_safe=True,
    # metadata to display on PyPI
    author="Federico Ravettino, Facundo Manuel Quiroga",
    author_email="facundoq@gmail.com",
    description="Invariance, Same-Equivariance and other measures for Neural Networks. Support for TensorFlow.",
    keywords="transformational measures equivariance same-equivariance invariance variance neural networks python pytorch numpy tensorflow",
    url=url,   # project home page, if any
    project_urls={
        "Bug Tracker": url+"/issues",
        "Documentation": url,
        "Source Code": url,
    },
    # check list at:
    # https://pypi.org/classifiers/
    classifiers=[
        "License :: OSI Approved :: Python Software Foundation License",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Visualization",
        
    ],

    # could also include long_description, download_url, etc.
    long_description=long_description,
    long_description_content_type='text/markdown',
    cmdclass={
        'upload': UploadCommand,
    },
)




from abc import ABC, abstractmethod
import tensorflow as tf
from dataset import DataSet


class TransformationsDataSet(DataSet):
    """
    A class used to have an abstract representation of a two dimensional transformations dataset

    ...

    Attributes
    ----------

    Methods
    -------


    """

    # abstract method
    def apply_inverse_transformation(self, tensor: tf.Tensor, index_transformation: int):
        pass

    # abstract method
    def get_size_transformations_groups(self):
        pass

    # abstract method
    def set_transformation_group(self, group: int, n: int):
        pass




from abc import ABC, abstractmethod


class DataSet(ABC):
    """
    A class used to represent a two dimensional dataset

    ...

    Attributes
    ----------

    Methods
    -------
    get_data_shape()
        Returns the shape of the data in the matrix.
    get_width()
        Returns the matrix width.
    get_height()
        Returns the matrix height.
    get_matrix(row:int,column:int)
        Returns a submatrix conformed by the rows and columns passed as arguments.
    transpose()
        Transpose the DataSet

    """
    # abstract method

    def get_data_shape(self):
        pass
    # abstract method

    def get_width(self):
        pass
    # abstract method

    def get_height(self):
        pass
    # abstract method

    def get_matrix(self, rows, columns):
        pass
    # abstract method

    def transpose(self):
        pass

    # abstract method
    def is_transpose(self):
        pass

    # abstract method
    def get_testing_data(self):
        pass
